@@ -3,7 +3,7 @@
  *
  * Copyright (C) 2005-2012  NTT DATA CORPORATION
  *
- * Version: 1.8.3+   2012/05/05
+ * Version: 1.8.5   2015/11/11
  */
 
 #include "internal.h"
@@ -27,17 +27,17 @@
 
 char *ccs_encode(const char *str);
 char *ccs_encode2(const char *str, int str_len);
-char *ccs_realpath(struct path *path);
+char *ccs_realpath(const struct path *path);
 const char *ccs_get_exe(void);
 void ccs_fill_path_info(struct ccs_path_info *ptr);
 
-static char *ccs_get_absolute_path(struct path *path, char * const buffer,
-				   const int buflen);
+static char *ccs_get_absolute_path(const struct path *path,
+				   char * const buffer, const int buflen);
 static char *ccs_get_dentry_path(struct dentry *dentry, char * const buffer,
 				 const int buflen);
 static char *ccs_get_local_path(struct dentry *dentry, char * const buffer,
 				const int buflen);
-static char *ccs_get_socket_name(struct path *path, char * const buffer,
+static char *ccs_get_socket_name(const struct path *path, char * const buffer,
 				 const int buflen);
 static int ccs_const_part_length(const char *filename);
 
@@ -213,8 +213,8 @@ static inline void ccs_realpath_unlock(void)
  *
  * If dentry is a directory, trailing '/' is appended.
  */
-static char *ccs_get_absolute_path(struct path *path, char * const buffer,
-				   const int buflen)
+static char *ccs_get_absolute_path(const struct path *path,
+				   char * const buffer, const int buflen)
 {
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 2, 0)
 	char *pos = ERR_PTR(-ENOMEM);
@@ -222,7 +222,7 @@ static char *ccs_get_absolute_path(struct path *path, char * const buffer,
 		pos = ccsecurity_exports.d_absolute_path(path, buffer,
 							 buflen - 1);
 		if (!IS_ERR(pos) && *pos == '/' && pos[1]) {
-			struct inode *inode = path->dentry->d_inode;
+			struct inode *inode = d_backing_inode(path->dentry);
 			if (inode && S_ISDIR(inode->i_mode)) {
 				buffer[buflen - 2] = '/';
 				buffer[buflen - 1] = '\0';
@@ -274,7 +274,7 @@ static char *ccs_get_absolute_path(struct path *path, char * const buffer,
 			ccs_no_empty = true;
 		}
 		if (!IS_ERR(pos) && *pos == '/' && pos[1]) {
-			struct inode *inode = path->dentry->d_inode;
+			struct inode *inode = d_backing_inode(path->dentry);
 			if (inode && S_ISDIR(inode->i_mode)) {
 				buffer[buflen - 2] = '/';
 				buffer[buflen - 1] = '\0';
@@ -293,7 +293,7 @@ static char *ccs_get_absolute_path(struct path *path, char * const buffer,
 		goto out;
 
 	*pos = '\0';
-	if (dentry->d_inode && S_ISDIR(dentry->d_inode->i_mode))
+	if (d_is_dir(dentry))
 		*--pos = '/';
 	for (;;) {
 		struct dentry *parent;
@@ -348,12 +348,10 @@ static char *ccs_get_dentry_path(struct dentry *dentry, char * const buffer,
 	if (buflen >= 256) {
 		/* rename_lock is locked/unlocked by dentry_path_raw(). */
 		pos = dentry_path_raw(dentry, buffer, buflen - 1);
-		if (!IS_ERR(pos) && *pos == '/' && pos[1]) {
-			struct inode *inode = dentry->d_inode;
-			if (inode && S_ISDIR(inode->i_mode)) {
-				buffer[buflen - 2] = '/';
-				buffer[buflen - 1] = '\0';
-			}
+		if (!IS_ERR(pos) && *pos == '/' && pos[1] &&
+		    d_is_dir(dentry)) {
+			buffer[buflen - 2] = '/';
+			buffer[buflen - 1] = '\0';
 		}
 	}
 	return pos;
@@ -362,7 +360,7 @@ static char *ccs_get_dentry_path(struct dentry *dentry, char * const buffer,
 	if (buflen < 256)
 		return ERR_PTR(-ENOMEM);
 	*pos = '\0';
-	if (dentry->d_inode && S_ISDIR(dentry->d_inode->i_mode))
+	if (d_is_dir(dentry))
 		*--pos = '/';
 	spin_lock(&dcache_lock);
 	while (!IS_ROOT(dentry)) {
@@ -425,13 +423,18 @@ static char *ccs_get_local_path(struct dentry *dentry, char * const buffer,
 	if (!MAJOR(sb->s_dev))
 		goto prepend_filesystem_name;
 	{
-		struct inode *inode = sb->s_root->d_inode;
+		struct inode *inode = d_backing_inode(sb->s_root);
 		/*
 		 * Use filesystem name if filesystems does not support rename()
 		 * operation.
 		 */
+#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 17, 0)
 		if (inode->i_op && !inode->i_op->rename)
 			goto prepend_filesystem_name;
+#else
+		if (!inode->i_op->rename && !inode->i_op->rename2)
+			goto prepend_filesystem_name;
+#endif
 	}
 	/* Prepend device name. */
 	{
@@ -473,10 +476,10 @@ out:
  *
  * Returns the buffer.
  */
-static char *ccs_get_socket_name(struct path *path, char * const buffer,
+static char *ccs_get_socket_name(const struct path *path, char * const buffer,
 				 const int buflen)
 {
-	struct inode *inode = path->dentry->d_inode;
+	struct inode *inode = d_backing_inode(path->dentry);
 	struct socket *sock = inode ? SOCKET_I(inode) : NULL;
 	struct sock *sk = sock ? sock->sk : NULL;
 	if (sk) {
@@ -501,7 +504,7 @@ static char *ccs_get_socket_name(struct path *path, char * const buffer,
  * This function uses kzalloc(), so caller must kfree() if this function
  * didn't return NULL.
  */
-char *ccs_realpath(struct path *path)
+char *ccs_realpath(const struct path *path)
 {
 	char *buf = NULL;
 	char *name = NULL;
@@ -533,12 +536,13 @@ char *ccs_realpath(struct path *path)
 			goto encode;
 		}
 #endif
-		inode = sb->s_root->d_inode;
+		inode = d_backing_inode(sb->s_root);
 		/*
 		 * Use local name for "filesystems without rename() operation"
 		 * or "path without vfsmount" or "absolute name is unavailable"
 		 * cases.
 		 */
+#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 17, 0)
 		if (!path->mnt || (inode->i_op && !inode->i_op->rename))
 			pos = ERR_PTR(-EINVAL);
 		else {
@@ -550,6 +554,14 @@ char *ccs_realpath(struct path *path)
 		if (pos == ERR_PTR(-EINVAL))
 			pos = ccs_get_local_path(path->dentry, buf,
 						 buf_len - 1);
+#else
+		if (!path->mnt ||
+		    (!inode->i_op->rename && !inode->i_op->rename2))
+			pos = ccs_get_local_path(path->dentry, buf,
+						 buf_len - 1);
+		else
+			pos = ccs_get_absolute_path(path, buf, buf_len - 1);
+#endif
 encode:
 		if (IS_ERR(pos))
 			continue;
@@ -703,23 +715,40 @@ void ccs_fill_path_info(struct ccs_path_info *ptr)
 const char *ccs_get_exe(void)
 {
 	struct mm_struct *mm = current->mm;
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 26)
 	struct vm_area_struct *vma;
-	const char *cp = NULL;
+#endif
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 20)
+	struct path path;
+#endif
+	struct file *exe_file = NULL;
+	const char *cp;
 	if (!mm)
 		return NULL;
 	down_read(&mm->mmap_sem);
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)
+	/* Not using get_mm_exe_file() as it is not exported. */
+	exe_file = mm->exe_file;
+#else
 	for (vma = mm->mmap; vma; vma = vma->vm_next) {
 		if ((vma->vm_flags & VM_EXECUTABLE) && vma->vm_file) {
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 20)
-			struct path path = { vma->vm_file->f_vfsmnt,
-					     vma->vm_file->f_dentry };
-			cp = ccs_realpath(&path);
-#else
-			cp = ccs_realpath(&vma->vm_file->f_path);
-#endif
+			exe_file = vma->vm_file;
 			break;
 		}
 	}
+#endif
+	if (exe_file)
+		get_file(exe_file);
 	up_read(&mm->mmap_sem);
+	if (!exe_file)
+		return NULL;
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 20)
+	cp = ccs_realpath(&exe_file->f_path);
+#else
+	path.mnt = exe_file->f_vfsmnt;
+	path.dentry = exe_file->f_dentry;
+	cp = ccs_realpath(&path);
+#endif
+	fput(exe_file);
 	return cp;
 }
